# -*- coding: utf-8 -*-
"""
Routines for plotting multiple sets.
"""
import numpy as np
import matplotlib.pyplot as plt

from supervenn._algorithms import (
    get_chunks_and_composition_array,
    get_permutations,
    DEFAULT_MAX_BRUTEFORCE_SIZE,
    DEFAULT_SEEDS,
    DEFAULT_NOISE_PROB)


DEFAULT_FONTSIZE = 12
DEFAULT_FIGSIZE = (8, 5)


def get_alternated_ys(ys_count, low, high):
    """
    A helper function generating y-positions for x-axis annotations, useful when some annotations positioned along the
    x axis are too crowded.
    :param ys_count: integer from 1 to 3.
    :param low: lower bound of the area designated for annotations
    :param high: higher bound for thr area designated for annotations.
    :return:
    """
    if ys_count not in [1, 2, 3]:
        raise ValueError('Argument ys_count should be 1, 2 or 3.')
    if ys_count == 1:
        coefs = [0.5]
        vas = ['center']
    elif ys_count == 2:
        coefs = [0.15, 0.85]
        vas = ['bottom', 'top']
    else:
        coefs = [0.15, 0.5, 0.85]
        vas = ['bottom', 'center', 'top']

    ys = [low + coef * (high - low) for coef in coefs]

    return ys, vas


def plot_binary_array(arr, ax=None, col_widths=None, row_heights=None, min_width_for_annotation=1,
                      row_annotations=None, row_annotations_y=0.5,
                      col_annotations=None, col_annotations_area_height=0.75, col_annotations_ys_count=1,
                      rotate_col_annotations=False,
                      color_by='row', bar_height=1, bar_alpha=0.6, bar_align='edge', color_cycle=None,
                      fontsize=DEFAULT_FONTSIZE):
    """
    Visualize a binary array as a grid with variable sized columns and rows, where cells with 1 are filled using bars
    and cells with 0 are blank.
    :param arr: numpy.array of zeros and ones
    :param ax: axis to plot into (current axis by default)
    :param col_widths: widths for grid columns, must have len equal to arr.shape[1]
    :param row_heights: heights for grid rows, must have len equal to arr.shape[0]
    :param row_annotations: annotations for each row, plotted in the middle of the row
    :param row_annotations_y: a number in (0, 1), position for row annotations in the row. Default 0.5 - center of row.
    :param col_annotations: annotations for columns, plotted in the bottom, below the x axis.
    :param col_annotations_area_height: height of area for column annotations in inches, 1 by default
    :param col_annotations_ys_count: 1 (default), 2, or 3 - use to reduce clutter in column annotations area
    :param rotate_col_annotations: True / False
    :param color_by: 'row' (default) or 'column'. If 'row', all cells in same row are same color, etc.
    :param bar_height: height of cell fill as a fraction of row height, a number in (0, 1).
    :param bar_alpha: alpha for cell fills.
    :param bar_align: vertical alignment of bars, 'edge' (defaulr) or 'center'. Only matters when bar_height < 1.
    :param color_cycle: a list of colors, given as names of matplotlib named colors, or hex codes (e.g. '#1f77b4')
    :param fontsize: font size for annotations (default {}).
    """.format(DEFAULT_FONTSIZE)
    if row_heights is None:
        row_heights = [1] * arr.shape[0]

    if col_widths is None:
        col_widths = [1] * arr.shape[1]

    if len(row_heights) != arr.shape[0]:
        raise ValueError('len(row_heights) doesnt match number of rows of array')

    if len(col_widths) != arr.shape[1]:
        raise ValueError('len(col_widths) doesnt match number of columns of array')

    allowed_argument_values = {
        'bar_align': ['center', 'edge'],
        'color_by': ['row', 'column'],
        'col_annotations_ys_count': [1, 2, 3],
    }

    for argument_name, allowed_argument_values in allowed_argument_values.items():
        if locals()[argument_name] not in allowed_argument_values:
            raise ValueError('Argument {} should be one of {}'.format(argument_name, allowed_argument_values))

    if not 0 <= row_annotations_y <= 1:
        raise ValueError('row_annotations_y should be a number between 0 and 1')

    if color_cycle is None:
        color_cycle = plt.rcParams['axes.prop_cycle'].by_key()['color']

    grid_xs = np.insert(np.cumsum(col_widths), 0, 0)[:-1]
    grid_ys = np.insert(np.cumsum(row_heights), 0, 0)[:-1]

    if ax is not None:
        plt.sca(ax)

    # BARS
    for row_index, (row, grid_y, row_height) in enumerate(zip(arr, grid_ys, row_heights)):

        bar_y = grid_y + 0.5 * row_height if bar_align == 'center' else grid_y

        # alternating background
        if row_index % 2:
            plt.barh(y=bar_y, left=0, width=sum(col_widths), height=bar_height * row_height, align=bar_align,
                     color='grey', alpha=0.15)

        for col_index, (is_filled, grid_x, col_width) in enumerate(zip(row, grid_xs, col_widths)):
            if is_filled:
                color_index = row_index if color_by == 'row' else col_index
                color = color_cycle[color_index % len(color_cycle)]
                plt.barh(y=bar_y, left=grid_x, width=col_width, height=bar_height * row_height, align=bar_align,
                         color=color, alpha=bar_alpha)

    # ROW ANNOTATIONS
    if row_annotations is not None:
        for row_index, (grid_y, row_height, annotation) in enumerate(zip(grid_ys, row_heights, row_annotations)):
            annot_y = grid_y + row_annotations_y * row_height
            plt.annotate(xy=(0.5 * sum(col_widths), annot_y), s=str(annotation),
                         ha='center', va='center', fontsize=fontsize)

    # COL ANNOTATIONS
    min_y = 0
    if col_annotations is not None:

        min_y = - 1.0 * col_annotations_area_height / plt.gcf().get_size_inches()[1] * arr.shape[0]
        plt.axhline(0, c='k')

        annot_ys, vas = get_alternated_ys(col_annotations_ys_count, min_y, 0)

        for col_index, (grid_x, col_width, annotation) in enumerate(zip(grid_xs, col_widths, col_annotations)):
            annot_y = annot_ys[col_index % len(annot_ys)]
            if col_width >= min_width_for_annotation:
                plt.annotate(xy=(grid_x + col_width * 0.5, annot_y), s=str(annotation),
                             ha='center', va=vas[col_index % len(vas)], fontsize=fontsize,
                             rotation=90 * rotate_col_annotations)

    plt.xlim(0, sum(col_widths))
    plt.ylim(min_y, sum(row_heights))
    plt.xticks(grid_xs, [])
    plt.yticks(grid_ys, [])
    plt.grid()


def side_plot(values, widths, orient, fontsize=DEFAULT_FONTSIZE, min_width_for_annotation=1, rotate_annotations=False,
              color='gray'):
    """
    Barplot with multiple bars of variable width right next to each other, with an option to rotate the plot 90 degrees.
    :param values: the values to be plotted.
    :param widths: Widths of bars
    :param orient: 'h' / 'horizontal' (default) or 'v' / 'vertical'
    :param fontsize: font size for annotations
    :param min_width_for_annotation: for horizontal plot, don't annotate bars of widths less than this value (to avoid
    clutter)
    """
    bar_edges = np.insert(np.cumsum(widths), 0, 0)
    annotation_positions = [0.5 * (begin + end) for begin, end in zip(bar_edges[:-1], bar_edges[1:])]
    max_value = max(values)
    if orient in ['h', 'horizontal']:
        horizontal = True
        plt.bar(x=bar_edges[:-1], height=values, width=widths, align='edge', alpha=0.5, color=color)
        ticks = plt.xticks
        noticks = plt.yticks
        lim = plt.ylim
    elif orient in ['v', 'vertical']:
        horizontal = False
        plt.barh(y=bar_edges[:-1], width=values, height=widths, align='edge', alpha=0.5, color=color)
        ticks = plt.yticks
        noticks = plt.xticks
        lim = plt.xlim
    else:
        raise ValueError('Unknown orient: {} (should be "h" or "v")'.format(orient))

    for i, (annotation_position, value, width) in enumerate(zip(annotation_positions, values, widths)):
        if width < min_width_for_annotation and horizontal:
            continue
        x, y = 0.5 * max_value, annotation_position
        if horizontal:
            x, y = y, x
        plt.annotate(xy=(x, y), s=value, ha='center', va='center', rotation=rotate_annotations * 90, fontsize=fontsize)

    ticks(bar_edges, [])
    noticks([])
    lim(0, max(values))
    plt.grid()


def balance_widths(widths, minmax_ratio=0.02):
    """
    Given a list of numbers, apply a linear transformation to every element of the array, such that the maximum value
    remains the same, and the minimum value is minmax_ratio times the maximum value
    :param widths: list of numbers
    :param minmax_ratio: the desired max / min ratio in the transformed list.
    :return: transformed list of numbers
    """
    if not 0 <= minmax_ratio <= 1:
        raise ValueError('minmax_ratio must be between 0 and 1')
    max_width = max(widths)
    min_width = min(widths)
    if 1.0 * min_width / max_width >= minmax_ratio:
        return widths
    slope = max_width * (1.0 - minmax_ratio) / (max_width - min_width)
    intercept = max_width * (max_width * minmax_ratio - min_width) / (max_width - min_width)
    return [slope * width + intercept for width in widths]


def supervenn(sets, set_annotations=None, figsize=DEFAULT_FIGSIZE, side_plots=True,
              chunks_ordering='minimize gaps', sets_ordering=None,
              reverse_chunks_order=True, reverse_sets_order=True,
              max_bruteforce_size=DEFAULT_MAX_BRUTEFORCE_SIZE, seeds=DEFAULT_SEEDS, noise_prob=DEFAULT_NOISE_PROB,
              side_plot_width=1.5, min_width_for_annotation=1, widths_minmax_ratio=0, side_plot_color='gray', **kw):
    """
    Plot a diagram visualizing relationship of multiple sets.
    :param sets: list of sets
    :param set_annotations: list of annotations for the sets
    :param figsize: figure size
    :param side_plots: True / False: add small barplots on top and on the right. On top, for each chunk it is shown,
    how many sets does this chunk lie inslde. On the right, set sizes are shown.
    :param chunks_ordering: method of ordering the chunks (columns of the grid)
        - 'minimize gaps' (default): use a smart algorithm to find an order of columns giving fewer gaps in each row,
            making the plot as readable as possible.
        - 'size': bigger chunks go first (or last if reverse_chunks_order=False)
        - 'occurence': chunks that are in most sets go first (or last if reverse_chunks_order=False)
        - 'random': randomly shuffle the columns
    :param sets_ordering: method of ordering the sets (rows of the grid)
        - None (default): keep the order as it is passed
        - 'minimize gaps': use a smart algorithm to find an order of rows giving fewer gaps in each column
        - 'size': bigger sets go first (or last if reverse_sets_order = False)
        - 'chunk count': sets that contain most chunks go first (or last if reverse_sets_order = False)
        - 'random': randomly shuffle
    :param reverse_chunks_order: True (default) / False when chunks_ordering is "size" or "occurence",
        chunks with bigger corresponding property go first if reverse_chunks_order=True, smaller go first if False.
    :param reverse_sets_order: True / False, works the same way as reverse_chunks_order
    :param max_bruteforce_size: maximal number of items for which bruteforce method is applied to find permutation
    :param seeds: number of different random seeds for the randomized greedy algorithm to find permutation
    :param noise_prob: probability of given element being equal to 1 in the noise array for randomized greedy algorithm
    :param side_plot_width: width of side plots in inches (default 1.5)
    :param side_plot_color: color of bars in side plots, default 'gray'
    :param min_width_for_annotation: for horizontal plot, don't annotate bars of widths less than this value (to avoid
    clutter)
    :param widths_minmax_ratio: desired max/min ratio of displayed chunk widths, default None (show actual widths)
    :param row_annotations_y: a number in (0, 1), position for row annotations in the row. Default 0.5 - center of row.
    :param col_annotations_area_height: height of area for column annotations in inches, 1 by default
    :param col_annotations_ys_count: 1 (default), 2, or 3 - use to reduce clutter in column annotations area
    :param color_by: 'row' (default) or 'column'. If 'row', all cells in same row are same color, etc.
    :param bar_height: height of cell fill as a fraction of row height, a number in (0, 1).
    :param bar_alpha: alpha for cell fills.
    :param bar_align: vertical alignment of bars, 'edge' (defaulr) or 'center'. Only matters when bar_height < 1.
    :param color_cycle: a list of colors, given as names of matplotlib named colors, or hex codes (e.g. '#1f77b4')
    """
    # Set up axes
    if side_plots:

        fig_width, fig_height = figsize

        height_ratios = [side_plot_width, fig_height - side_plot_width]
        width_ratios = [fig_width - side_plot_width, side_plot_width]

        _, axes = plt.subplots(2, 2, figsize=figsize, gridspec_kw={'height_ratios': height_ratios,
                                                                   'width_ratios': width_ratios,
                                                                   'hspace': 0.0,
                                                                   'wspace': 0.0})
        for side_ax in axes[0, 1], axes[1, 1], axes[0, 0]:
            side_ax.set_xticks([])
            side_ax.set_yticks([])
        main_ax = axes[1, 0]
    else:
        plt.figure(figsize=figsize)
        main_ax = plt.gca()

    if set_annotations is None:
        set_annotations = ['Set_{}'.format(i) for i in range(len(sets))]

    chunks, composition_array = get_chunks_and_composition_array(sets)

    # Find permutations of rows and columns
    permutations_ = get_permutations(
        chunks,
        composition_array,
        chunks_ordering=chunks_ordering,
        sets_ordering=sets_ordering,
        reverse_chunks_order=reverse_chunks_order,
        reverse_sets_order=reverse_sets_order,
        max_bruteforce_size=max_bruteforce_size,
        seeds=seeds,
        noise_prob=noise_prob)

    # Apply permutations
    chunks = [chunks[i] for i in permutations_['chunks_ordering']]
    composition_array = composition_array[:, permutations_['chunks_ordering']]
    composition_array = composition_array[permutations_['sets_ordering'], :]
    set_annotations = [set_annotations[i] for i in permutations_['sets_ordering']]

    # Main plot
    chunk_sizes = [len(chunk) for chunk in chunks]

    col_widths = balance_widths(chunk_sizes, widths_minmax_ratio) if widths_minmax_ratio is not None else chunk_sizes

    plot_binary_array(
        arr=composition_array,
        row_annotations=set_annotations,
        col_annotations=chunk_sizes,
        ax=main_ax,
        col_widths=col_widths,
        row_heights=[1] * len(sets),
        min_width_for_annotation=min_width_for_annotation,
        **kw)

    xlim = main_ax.get_xlim()
    ylim = main_ax.get_ylim()
    plt.xlabel('ITEMS', fontsize=kw.get('fontsize', DEFAULT_FONTSIZE))
    plt.ylabel('SETS', fontsize=kw.get('fontsize', DEFAULT_FONTSIZE))

    # Side plots
    if side_plots:
        plt.sca(axes[0, 0])
        side_plot(composition_array.sum(0), col_widths, 'h', min_width_for_annotation=min_width_for_annotation,
                  rotate_annotations=kw.get('rotate_col_annotations', False), color=side_plot_color)
        plt.xlim(xlim)

        plt.sca(axes[1, 1])
        side_plot([len(sets[i]) for i in permutations_['sets_ordering']], [1] * len(sets), 'v', color=side_plot_color)
        plt.ylim(ylim)



# -*- coding: utf-8 -*-
"""
    supervenn._algorithms
    This module implements all the algorithms used to prepare data for plotting.
    ~~~~~~~~~~~
    A semi-formal explanation of what is going on here
    ==================================================

    CHUNKS

    Consider a three-way Venn diagram, shown below using squares rather than the usual circles, for obvious reasons.
    It is easy to see that the three squares S1, S2, S3 break each other into seven elementary undivided parts, which we
    will name :chunks:. The 7 chunks are marked as ch1 - ch7.


              *-------S1-------*
              |                |
              |           ch1  |
        *-----|---------*      |
        |     |  ch4    |      |
        |     |         |      |
        |     |   *-----|------|-----*
        |     |   | ch7 | ch5  |     |
       S2     |   |     |      |     |
        |     *---|-----|------*     |
        |         |     |            |
        |  ch2    | ch6 |            S3
        *---------|-----*            |
                  |          ch3     |
                  |                  |
                  *------------------*

    The number 7 is easily derived by the following simple consideration. Each chunk is defined by a unique
    subset of {S1, S2, S3}. For example, chunk ch4 is defined by {S1, S2} - it consists of exactly the points that lie
    inside S1 and S2, but outside of S3. So the number of chunks is 2^3 - 1, 2^3 being the number of possible subsets of
    {S1, S2, S3}, and -1 being for the empty set which has no business being among our chunks.

    Note that, depending on how the three squares are positioned, there can be less chunks then 7. For instance, if the
    three squares are disjoint, the number of chunks will be 3, as each square will be a chunk in itself.

    What is important for us about the chunks, is that they are like the elementary building blocks of our
    configuration of squares. In other words, we can represent any of the squares as a disjoint union of some of the
    chunks. Also any combination of the squares w.r.t. the set-theoretical operations of intersection, union and
    difference can be represented in the same way.

    Now let's move from squares on the plane to abstract finite sets. The same line of reasoning leads us to
    the conclusion that for N sets S1, ..., SN, there exist at most 2^N - 1 chunks (sets), such that any of the sets
    (and any combination of the sets w.r.t union, intersection and difference) can be uniquely represented as the
    disjoint union of several of the chunks.

    COMPOSITION ARRAY

    Suppose we have N sets S1, ..., SN, and we have found the chunks ch1, ... chM. As we said above, each of the sets
    can be uniquely represented as a disjoint union of some of the chunks. Suppose we have found such decompositions for
    all of the sets.
    Now, represent the decompositions as a N x M array of zeros and ones, where the i,j-th item is 1 <=> set S_i
    contains chunk ch_j. We'll call this array the :composition array: of our sets.
    For example, for the squares above, the composition array will look like this:

       ch1  ch2  ch3  ch4  ch5  ch6  ch7
    S1  1    0    0    1    1    0    1  ->  Square 1 is made up of chunks ch1, ch4, ch5 and ch7
    S2  0    1    0    1    0    1    1
    S3  0    0    1    0    1    1    1

    Note that since chunks are something we've just made up by ourselves, they don't have to be ordered in any
    particular way. This means that we can reorder the chunks (= the columns if the composition array) if we
    need to.

    Most part of the present module is actually about finding a permutations of the columns, that will minimize the
    number of gaps between the 1's in all the rows of the array, so that the sets are visually broken into as few parts
    as possible.

    For the array above, we see that there are 5 total zero-filled gaps between ones: one in the first row, two in the
    second, one in the third. But if we find and apply the right permutation of columns, we'll have only one gap between
    the 1's in the whole array, highlighted by asterisks below:


       ch1  ch4  ch2  ch6  ch7  ch5  ch3
    S1  1    1   *0*   1    1    0    0
    S2  0    1    1    1    1    0    0
    S3  0    0    0    1    1    1    1

    When there are few chunks (say, no more than 8) the optimal permutations is easily found in a bruteforce manner by
    checking all the possible permutations. But for a greater number of chunks this would take too long. So instead an
    approximate greedy algorithm with randomization is used. It is implemented in the run_randomized_greedy_algorithm()
    function. The algorithm will be described below.

    REORDERING THE ROWS / SETS

    Unlike the order of chunks, the order of sets can be relevant in some cases. For example, our sets sets may relate
    to different moments in time, and we may deem it important to preserve this temporal structure in the plot. But in
    other cases when there is no inherent order to our sets, we might want to reorder the sets (rows of the composition
    array) so that similar sets are closer to each other.

    This is done by running the same algorithm on the transposed array. But there is an important distinction here: the
    chunks are of different size, and it makes more sense to minimize total gap width, instead of just gap count.
    To allow that, the run_randomized_greedy_algorithm() function accepts an additional argument named :row_weights:,
    and minimizes gap counts weighted with these coefficients. In other words, the minimization target becomes
    sum_i(gaps_count_in_row[i] * row_weights[i]) instead of just sum_i(gaps_count_in_row[i]).

    DESCRIPTION OF THE ALGORITHM

    Recall that our goal is to a find a permutation of columns of a matrix of zeros and ones, so that the row-weighted
    sum of gaps counts in rows is as low as possible.

    Define similarity of two columns as the sum of row weights for rows where the values are equal in the two columns.
    Precompute matrix of similarities of all columns (if arr has N columns, this matrix has shape N x N)
    Find two most similar columns C1, C2, initialize two lists [C1,] and [C2,]
    Among the remaining N-2 columns, find one that has largest similarity to either of C1 or C2. Append that column to
    the corresponding list. E.g. we now have lists [C1, C3] and [C2,]
    Among the remaining N-3 columns find one that has largest similarity to the last elemt of either of the lists (in
    our example, C3 and C2). Append it to the corresponding list.
    Continue until all columns are distributed between in two lists.
    Finally, one of the lists is reversed and the other is concatenated to it on the right, which gives the resulting
    permutation. This concludes the greedy algorithm.

    To mitigate the greediness, the similarity matrix is perturbed by means of adding a random matrix R with
    elements {0, noise_value} with a fixed  probability of non-zero, and the whole greedy procedure is repeated.
    This is done with :seeds: different random matrices, and so :seeds: + 1 permutations are obtained (including the one
    produced from the unperturbed similarities matrix). The permutation that yields the smallest value
    of count_runs_of_ones_in_rows(arr[:, permutation]) is returned.
"""
from collections import defaultdict
import datetime
from itertools import permutations

import numpy as np

HUGE_NUMBER = 1e10 # can fail for weighted! FIXME
DEFAULT_MAX_BRUTEFORCE_SIZE = 8
BRUTEFORCE_SIZE_HARD_LIMIT = 12
DEFAULT_SEEDS = 1000
DEFAULT_NOISE_PROB = 0.0075
DEFAULT_MAX_NOISE = 1.1


def get_total_gaps_in_rows(arr, row_weights=None):
    """
    In a numpy.array arr, count how many gaps of zeros are there between contigous runs of non-zero values in each row.
    The counts in each row are multiplied by weights given by row_weights array and summed. By default, all row weights
    are equal to 1, so the returned value is just the total count of gaps in rows.
    :param arr: 2D numpy.array, all that matters is which elements are zero and which are not.
    :param row_weights: 1D numpy array with len equal to len(arr), provides weights for each row.
    :return: weighted sum of number of gaps in rows.
    """
    if row_weights is None:
        row_weights = np.ones(len(arr), dtype=int)
    if len(arr) != len(row_weights):
        raise ValueError('len(row_weights) == {} != {} == len(arr)'.format(len(row_weights), len(arr)))

    arr = arr.astype(bool)

    # Shift array to the left, dropping last column and appending column of zeros on the left.
    shifted_arr = np.concatenate([np.zeros(len(arr), dtype=bool).reshape((-1, 1)), arr[:, :-1]], 1)
    rowwise_runs_counts = (arr & (~shifted_arr)).sum(1)
    rowwise_gaps_counts = np.maximum(rowwise_runs_counts - 1, 0)

    return rowwise_gaps_counts.dot(row_weights)


def get_chunks_and_composition_array(sets):
    """
    Let us have a collection {S_1, ..., S_n} of finite sets and U the union of all these sets.
    For a given subset C = {i_1, ..., i_k} of indices {1, ..., n}, define the 'chunk', corresponding to C, as the set
    of elements of U, that belong to S_i_1, ..., S_i_k, but not to any of the other sets.
    For example, for a collection of two sets {S_1, S_2}, there can be max three chunks: S_1 - S_2, S_2 - S_1, S1 & S_2.
    For three sets, there can be max 7 chunks (imagine a generic three-way Venn diagram and count how many different
    area colors it can have).
    In general, the number of possible non-empty chunks for a collection of n sets is equal to min(|U|, 2^n - 1).
    Any chunk either lies completely inside any or completely outside any of the sets S_1, ... S_n.

    This function takes a list of sets as its only argument and returns two objects:
    - list of all chunks (each chunk is a set of items)
    - a numpy.array A of zeros and ones with len(sets) rows and len(chunks) columns,
    where A[i, j] == 1 <=> sets[i] includes chunks[j].
    :param sets: list of sets
    :return: chunks - list of sets, arr - numpy.array
    """

    if not sets:
        raise ValueError('Sets list is empty.')

    all_items = set.union(*sets)

    if not all_items:
        raise ValueError('All sets are empty')

    # Each chunk is characterized by its occurence pattern, which is a unique subset of indices of our sets.
    # E.g. chunk with signature {1, 2, 5} is exactly the set of items such that they belong to sets 1, 2, 5, and
    # don't belong to any of the other sets.
    # Build a dict with signatures as keys (as frozensets), and lists of items as values,
    occurences = defaultdict(set)
    for item in all_items:
        occurence_pattern = frozenset({i for i, set_ in enumerate(sets) if item in set_})
        occurences[occurence_pattern].add(item)

    chunks_count = len(occurences)
    chunks = []
    arr = np.zeros((len(sets), chunks_count), dtype=int)

    for idx, (sets_indices, items) in enumerate(occurences.items()):
        chunks.append(items)
        arr[list(sets_indices), idx] = 1

    return chunks, arr


def find_best_columns_permutation_bruteforce(arr, row_weights=None):
    """
    Using exhaustive search, find permutation of columns of np.array arr that will provide minimum value of weighted sum
    of gaps counts in rows of arr. Will take unreasonably long time if arr has > 10 cols.
    :param arr: 2D numpy.array.  All that matters is which elements are zero and which are not
    :param row_weights: 1D numpy array with len equal to len(arr). Provides weights for each row.
    :return: optimal permutation as a list of column indices.
    """
    if arr.shape[1] > BRUTEFORCE_SIZE_HARD_LIMIT:
        raise ValueError('Bruteforce ordering method accepts max {} columns, got {} instead. It would take too long.'
                         .format(BRUTEFORCE_SIZE_HARD_LIMIT, arr.shape[1]))
    best_permutation = None
    best_gaps_count = HUGE_NUMBER
    for permutation in permutations(range(arr.shape[1])):
        gaps_count = get_total_gaps_in_rows(arr[:, permutation], row_weights=row_weights)
        if gaps_count < best_gaps_count:
            best_permutation = permutation
            best_gaps_count = gaps_count
    return list(best_permutation)


def columns_similarities_matrix(arr, row_weights=None):
    """
    Let A and B be two 1D arrays with zeros and ones, of same shape, and row_weigths a non-negative array of same shape.
    Define weighted similarity of A and B w.r.t. row weights as the sum of elements of row_weights in positions where
    A and B have equal values.
    So, given a M x N array arr, this function computes the N x N matrix of weighted similarities of its columns.
    All weights are equal to 1 by default, in which case the similarity of two columns is equal to the number of
    positions where values coincide in the two columns.
    :param arr: M x N array with zeros and ones
    :param row_weights: 1D numpy array with len equal to len(arr). Provides weights for each row.
    :return: N x N array. with weighted similarities of columns of arr as defined above.
    """

    if row_weights is None:
        row_weights = np.ones(len(arr), dtype=int)
    if len(arr) != len(row_weights):
        raise ValueError('len(row_weights) must be equal to number of rows of arr')

    return (arr.T * row_weights).dot(arr) + ((1 - arr).T * row_weights).dot(1 - arr)


def find_columns_permutation_greedily(similarities):
    """
    Given an array of columns similarities, use a greedy algorithm to try and find a permutation of columns that
    will lower the number of gaps in rows.
    :param similarities: numpy.array of column similarities.
    :return: a permutation (as a list of indices) of column indices, producing lower value of count_runs
    """

    if len(similarities) == 1:
        return [0]

    # fill diagonal with negative value so that same column is never paired with itself, even if all similarities
    # between different columns are zero.
    similarities = similarities.copy()
    np.fill_diagonal(similarities, -1)

    ncols = similarities.shape[0]

    # define placed_flags[i] == 1 <=> i-th column of arr has already been assinged a place in the permutation
    placed_flags = np.zeros(ncols, dtype=int)

    # find two most similar columns. Initialize two lists with tham.
    first_col_index, second_col_index = np.unravel_index(similarities.argmax(), similarities.shape)
    first_tail = [first_col_index]
    second_tail = [second_col_index]
    placed_flags[first_col_index] = 1
    placed_flags[second_col_index] = 1

    # the main greedy loop
    for _ in range(ncols - 2):
        # find column most similar to the last element of any of the two lists, append it to the corresponding list
        similarities_to_first = similarities[:, first_tail[-1]] - placed_flags * HUGE_NUMBER
        similarities_to_second = similarities[:, second_tail[-1]] - placed_flags * HUGE_NUMBER
        first_argmax = similarities_to_first.argmax()
        second_argmax = similarities_to_second.argmax()
        if similarities_to_first[first_argmax] >= similarities_to_second[second_argmax]:
            idx = first_argmax
            first_tail.append(idx)
        else:
            idx = second_argmax
            second_tail.append(idx)

        # mark the chosen column as placed
        placed_flags[idx] = 1

    permutation = second_tail[::-1] + first_tail

    return permutation


# todo rename to reflect what it does, not only how it does it
def run_greedy_algorithm_on_composition_array(arr, row_weights=None):
    """
    Given a composition array, use a greedy algorithm to find a permutation of columns that will try to minimize the
    total weighted number of gaps in the rows of the array.
    :param arr: numpy array with zeros and ones
    :param row_weights: 1D numpy array with len equal to len(arr). Provides weights for each row.
    :return:
    """
    similarities = columns_similarities_matrix(arr, row_weights=row_weights)
    return find_columns_permutation_greedily(similarities)


# todo rename to reflect what it does, not only how it does it
def run_randomized_greedy_algorithm(arr, row_weights=None, seeds=DEFAULT_SEEDS, noise_prob=DEFAULT_NOISE_PROB):
    """
    For a 2D np.array arr, find a permutation of columns that approximately minimizes the row-weighted number of gaps in
    the rows of the permuted array. An approximate randomized greedy algorithm is used.



    :param arr: np.array
    :param seeds: number of different random perturbations to figh sticking in local minima.
    :return: a list of indices representing an approximately optimal permutation of columns of arr
    """

    arr = arr.astype(int)

    # Compute similarities matrix
    similarities = columns_similarities_matrix(arr, row_weights=row_weights)

    best_found_permutation = None
    best_found_gaps_count = HUGE_NUMBER

    for seed in range(seeds + 1):
        # Perturb similarities matrix if seed != 0
        if seed == 0:
            noise = np.zeros_like(similarities)
        else:
            np.random.seed(seed)
            noise = (np.random.uniform(0, 1, size=similarities.shape) < noise_prob).astype(int)
            np.random.seed(datetime.datetime.now().microsecond)

        np.fill_diagonal(noise, 0)
        noisy_similarities = similarities + noise

        permutation = find_columns_permutation_greedily(noisy_similarities)

        # Compute the target value for the found permutation, compare with current best.
        gaps = get_total_gaps_in_rows(arr[:, permutation], row_weights=row_weights)
        if gaps < best_found_gaps_count:
            best_found_permutation = permutation
            best_found_gaps_count = gaps

    return best_found_permutation


def get_permutations(chunks, composition_array, chunks_ordering='minimize gaps', sets_ordering=None,
                     reverse_chunks_order=True, reverse_sets_order=True,
                     max_bruteforce_size=DEFAULT_MAX_BRUTEFORCE_SIZE,
                     seeds=DEFAULT_SEEDS, noise_prob=DEFAULT_NOISE_PROB):
    """
    Given chunks and composition array, get permutations which will order the chunks and the sets according to specified
    ordering methods.
    :param chunks, composition_array: as returned by get_chunks_and_composition_array
    For explanation of other params, see docstring to supervenn()
    :return: a dict of the form {'chunks_ordering': [3, 2, 5, 4, 1, 6, 0], 'sets_ordering': [2, 0, 1, 3]}
    """
    chunk_sizes = [len(chunk) for chunk in chunks]
    set_sizes = composition_array.dot(np.array(chunk_sizes).reshape((-1, 1)))

    chunks_case = {'sizes': chunk_sizes,
                   'param': 'chunks_ordering',
                   'array': composition_array,
                   'row_weights': None,
                   'ordering': chunks_ordering,
                   'allowed_orderings': ['size', 'occurence', 'random', 'minimize gaps'],
                   'reverse': reverse_chunks_order}

    sets_case = {'sizes': set_sizes,
                 'param': 'sets_ordering',
                 'array': composition_array.T,
                 'row_weights': chunk_sizes,
                 'ordering': sets_ordering,
                 'allowed_orderings': ['size', 'chunk count', 'random', 'minimize gaps', None],
                 'reverse': reverse_sets_order}

    permutations_ = {}

    for case in chunks_case, sets_case:
        if case['ordering'] not in case['allowed_orderings']:
            raise ValueError('Unknown {}: {} (should be one of {})'
                             .format(case['param'], case['ordering'], case['allowed_orderings']))

        if case['ordering'] == 'size':
            permutation = np.argsort(case['sizes'])
        elif case['ordering'] in ['occurence', 'chunk count']:
            permutation = np.argsort(case['array'].sum(0))
        elif case['ordering'] == 'random':
            permutation = np.array(range(len(case['sizes'])))
            np.random.shuffle(permutation)
        elif case['ordering'] is None:
            permutation = np.array(range(len(case['sizes'])))
        elif case['ordering'] == 'minimize gaps':
            if len(case['sizes']) <= min(max_bruteforce_size, BRUTEFORCE_SIZE_HARD_LIMIT):
                permutation = find_best_columns_permutation_bruteforce(case['array'], row_weights=case['row_weights'])
            else:
                permutation = run_randomized_greedy_algorithm(case['array'], seeds=seeds, noise_prob=noise_prob,
                                                              row_weights=case['row_weights'])
        else:
            raise ValueError(case['ordering'])

        if case['ordering'] is not None and case['reverse']:
            permutation = permutation[::-1]

        permutations_[case['param']] = permutation

    return permutations_


def _get_ordered_chunks_and_composition_array(sets, **kwargs):
    """
    Wrapper for get_permutations, used only for testing
    :param sets: list of sets
    :param kwargs: all arguments to get_permutations except for chunks and composition_array,
    :return:
    """
    chunks, composition_array = get_chunks_and_composition_array(sets)

    permutations_ = get_permutations(chunks, composition_array, **kwargs)

    ordered_chunks = [chunks[i] for i in permutations_['chunks_ordering']]
    ordered_composition_array = composition_array[:, permutations_['chunks_ordering']]
    ordered_composition_array = ordered_composition_array[permutations_['sets_ordering'], :]

    return ordered_chunks, ordered_composition_array



# -*- coding: utf-8 -*-
"""
Tests for _algorithms module.
"""
import unittest
import numpy as np
from collections import Counter
from itertools import product


from supervenn._algorithms import (
    get_total_gaps_in_rows,
    get_chunks_and_composition_array,
    _get_ordered_chunks_and_composition_array,
    find_best_columns_permutation_bruteforce,
    run_greedy_algorithm_on_composition_array,
    run_randomized_greedy_algorithm
)


def freeze_sets(sets):
    """
    Convert an iterable of sets to a set of frozensets, and check that all sets are not repeated.
    """
    counter = Counter(frozenset(set_) for set_ in sets)
    assert max(counter.values()) == 1
    return set(counter)


def is_ascending(sequence):
    return all(sequence[i + 1] >= sequence[i] for i in range(len(sequence) - 1))


def array_is_binary(arr):
    """
    :param arr: np.array
    :return: True if the array only contains zeros and ones, False otherwise
    """
    return not set(arr.flatten()) - {0, 1}


def make_random_sets(min_sets_count=1):
    sets_count = np.random.randint(min_sets_count, 10)
    max_item = np.random.choice([2 ** i for i in range(8)]) + 1
    max_size = np.random.choice([2 ** i for i in range(8)]) + 1
    sets = [set(np.random.randint(1, max_item, size=np.random.randint(1, max_size)))
            for _ in range(sets_count)]

    # decimate sets to introduce some empty sets into the list
    for index in np.random.randint(0, len(sets), size=int(len(sets) / 10)):
        sets[index] = set()
    return sets


def is_permutation(sequence):
    n = len(sequence)
    if len(sequence) != n:
        return False
    if len(set(sequence)) != n:
        return False
    if set(sequence) != set(range(n)):
        return False
    return True


class TestGetTotalGaps(unittest.TestCase):

    def test_all_zeros(self):
        """
        Test that number of runs of a zero matrix is zero
        """
        arr = np.zeros((2, 3), dtype=bool)
        gaps_count = get_total_gaps_in_rows(arr)
        self.assertEqual(gaps_count, 0)

    def test_all_ones(self):
        """
        Test that the number of runs of a matrix of ones is equal to the number of rows.
        """
        arr = np.ones((7, 3), dtype=bool)
        gaps_count = get_total_gaps_in_rows(arr)
        self.assertEqual(gaps_count, 0)

    def test_one_col(self):
        """
        Test that the number of runs of a matrix with one col is equal to the number of ones
        """
        rows_count = 100
        arr = np.random.randint(0, 2, size=(rows_count, 1))
        gaps_count = get_total_gaps_in_rows(arr)
        self.assertEqual(gaps_count, 0)

    def test_nontrivial_one(self):
        arr = np.array([[1, 0, 1, 0],
                        [1, 0, 0, 1],
                        [0, 1, 0, 1]])
        gaps_count = get_total_gaps_in_rows(arr)
        self.assertEqual(gaps_count, 3)

    def test_nontrivial_two(self):
        arr = np.array([[1, 0, 0, 0],
                        [0, 0, 0, 1],
                        [0, 0, 0, 1]])
        gaps_count = get_total_gaps_in_rows(arr)
        self.assertEqual(gaps_count, 0)

    def test_nontrivial_three(self):
        arr = np.array([[1, 1, 0, 1],
                        [0, 1, 1, 0],
                        [1, 0, 1, 1]])
        gaps_count = get_total_gaps_in_rows(arr)
        self.assertEqual(gaps_count, 2)


class TestGetChunksAndCompositionArray(unittest.TestCase):
    def test_single_set(self):
        sets = [set(np.random.randint(1, 1000, size=100))]
        chunks, arr = get_chunks_and_composition_array(sets)
        self.assertEqual(chunks, sets)
        self.assertEqual(arr.shape, (1, 1))
        self.assertEqual(arr[0, 0], 1)

    def test_one_empty(self):
        set_ = set(np.random.randint(1, 1000, size=100))
        sets = [set_, set()]
        chunks, arr = get_chunks_and_composition_array(sets)
        self.assertEqual(chunks, [set_])
        self.assertEqual(arr.shape, (2, 1))
        self.assertEqual(list(arr[:, 0]), [1, 0])

    def test_disjoint_small(self):
        sets = [{1}, {2}]
        chunks, arr = get_chunks_and_composition_array(sets)
        self.assertEqual(freeze_sets(sets), freeze_sets(chunks))
        self.assertTrue(np.array_equal(arr, np.eye(2, dtype=int)))  # fixme can be not eye

    def test_disjoint_large(self):
        set_count = 4
        sets = [set(np.random.randint(1000 * i, 1000 * (i + 1), size=100 * (i + 1))) for i in range(set_count)]
        chunks, arr = get_chunks_and_composition_array(sets)
        self.assertEqual(freeze_sets(sets), freeze_sets(chunks))
        # Verify that array is made of zeros and ones
        self.assertTrue(array_is_binary(arr))
        # Verify that each row and each column of arr sum to 1
        self.assertEqual(set(arr.sum(0)), {1})
        self.assertEqual(set(arr.sum(1)), {1})

    def test_two_sets(self):
        sets = [{1, 2, 3}, {3, 4}]
        chunks, arr = get_chunks_and_composition_array(sets)
        self.assertEqual(freeze_sets(chunks), {frozenset([1, 2]), frozenset([3]), frozenset([4])})
        self.assertTrue(array_is_binary(arr))

    def test_chunks_for_random_sets(self):
        """
        For random sets, test that
        1) the union of sets is the disjoint union of chunks
        2) each chunk is either completely inside any set either completely outside
        :return:
        """
        for _ in range(100):
            sets = make_random_sets()
            chunks, arr = get_chunks_and_composition_array(sets)
            all_elements = set.union(*sets)
            self.assertEqual(set.union(*chunks), all_elements)
            self.assertEqual(sum(len(chunk) for chunk in chunks), len(all_elements))
            for chunk, set_ in product(chunks, sets):
                self.assertTrue(not chunk - set_ or chunk - set_ == chunk)

    def test_chunks_and_array(self):
        """
        For random sets, test that the you indeed can recreate every original set as the union of chunks indicated in
        the corresponding row of array.
        """
        for _ in range(100):
            sets = make_random_sets()
            chunks, arr = get_chunks_and_composition_array(sets)
            for (set_, row) in zip(sets, arr):
                chunks_in_this_set = [chunk for is_included, chunk in zip(row, chunks) if is_included]
                recreated_set = set.union(*chunks_in_this_set)
                self.assertEqual(set_, recreated_set)
                self.assertEqual(sum(len(chunk) for chunk in chunks_in_this_set), len(set_))


def make_test_for_algorithm(function):

    class TestPermutationFinder(unittest.TestCase):
        def test_single_set(self):
            arr = np.array([[1]])
            permutation = function(arr)
            self.assertEqual(permutation, [0])

        def test_single_chunk(self):
            arr = np.array([1, 0, 1, 0]).reshape((-1, 1))
            permutation = function(arr)
            self.assertEqual(permutation, [0])

        def test_two_cols(self):
            for sets_count in 1, 2, 10:
                col_1 = (np.random.uniform(0, 1, size=sets_count) > 0.6).astype(int)
                col_2 = (np.random.uniform(0, 1, size=sets_count) > 0.6).astype(int)
                col_1[0] = 1
                col_2[0] = 1
                arr = np.concatenate([col_1.reshape((-1, 1)), col_2.reshape((-1, 1))], 1)
                permutation = function(arr)
                self.assertIn(permutation, [[0, 1], [1, 0]])

        def test_is_permutation(self):
            for _ in range(100):
                ncols = np.random.randint(1, 5)
                nrows = np.random.randint(1, 5)
                arr = (np.random.uniform(0, 1, size=(nrows, ncols)) > 0.5).astype(int)
                arr[0][0] = 1
                permutation = function(arr)
                self.assertTrue(is_permutation(permutation))

        def test_obvious_one(self):
            arr = np.array([[1, 1, 1, 1],
                            [1, 0, 1, 0],
                            [1, 0, 1, 0]])
            permutation = function(arr)
            permutation_string = ''.join((str(i) for i in permutation))
            self.assertIn('02', permutation_string + ' ' + permutation_string[::-1])

        def test_obvious_two(self):
            arr = np.array([[0, 1, 0, 1, 0, 0],
                            [1, 0, 1, 1, 0, 1],
                            [1, 0, 0, 1, 1, 0]])
            permutation = function(arr)
            permutation_string = ''.join((str(i) for i in permutation))
            self.assertIn('03', permutation_string + ' ' + permutation_string[::-1])

        def test_obvious_three(self):
            arr = np.array([[0, 1, 0, 1, 1, 0],
                            [1, 0, 1, 1, 0, 1],
                            [1, 0, 0, 1, 0, 0],
                            [1, 1, 0, 1, 0, 0],
                            [1, 0, 0, 0, 1, 0]])
            permutation = function(arr)
            permutation_string = ''.join((str(i) for i in permutation))
            self.assertIn('03', permutation_string + ' ' + permutation_string[::-1])
            self.assertIn('13', permutation_string + ' ' + permutation_string[::-1])

    return TestPermutationFinder


TestFindBestColumnPermutationBruteforce = make_test_for_algorithm(find_best_columns_permutation_bruteforce)
TestRunGreedyAlgorithmOnCompositionArray = make_test_for_algorithm(run_greedy_algorithm_on_composition_array)
TestRunRandomizedGreedyAlgorithm = make_test_for_algorithm(run_randomized_greedy_algorithm)

# TODO: A test that randomized is not worse than non-randomized

class TestRandomizedAlgorithmQuality(unittest.TestCase):

    def test_quality(self):
        for _ in range(100):
            one_prob = np.random.uniform(0.1, 0.9)
            nrows = np.random.randint(1, 30)
            ncols = np.random.randint(1, 8)
            arr = np.random.uniform(0, 1, size=(nrows, ncols)) < one_prob
            permutation = run_randomized_greedy_algorithm(arr)
            target = get_total_gaps_in_rows(arr[:, permutation])
            best_permutation = find_best_columns_permutation_bruteforce(arr)
            best_target = get_total_gaps_in_rows(arr[:, best_permutation])
            self.assertLessEqual((target - best_target), 0.3 * best_target + 1)


class TestRandomizedAlgorithmReproducible(unittest.TestCase):

    def test_reproducible(self):
        arr = np.random.uniform(0, 1, size=(20, 20)) < 0.5
        representations_set = set()
        for _ in range(10):
            permutation = run_randomized_greedy_algorithm(arr)
            representations_set.add(str(permutation))
        self.assertEqual(len(representations_set), 1)


class TestGetOrderedChunksAndCompositionArray(unittest.TestCase):
    def test_order_chunks_size_descending(self):
        for _ in range(10):
            sets = make_random_sets()
            chunks, _ = _get_ordered_chunks_and_composition_array(sets, chunks_ordering='size')
            chunk_sizes_descending = is_ascending([len(chunk) for chunk in chunks][::-1])
            self.assertTrue(chunk_sizes_descending)

    def test_order_chunks_size_ascending(self):
        for _ in range(10):
            sets = make_random_sets()
            chunks, _ = _get_ordered_chunks_and_composition_array(sets, chunks_ordering='size',
                                                                  reverse_chunks_order=False)
            chunk_sizes_ascending = is_ascending([len(chunk) for chunk in chunks])
            self.assertTrue(chunk_sizes_ascending)

    def test_order_chunks_occurence_descending(self):
        for _ in range(10):
            sets = make_random_sets()
            _, composition_matrix = _get_ordered_chunks_and_composition_array(sets, chunks_ordering='occurence')
            occurences = composition_matrix.sum(0)
            occurences_descending = is_ascending(occurences[::-1])
            self.assertTrue(occurences_descending)

    def test_order_chunks_occurence_ascending(self):
        for _ in range(10):
            sets = make_random_sets()
            _, composition_matrix = _get_ordered_chunks_and_composition_array(sets, chunks_ordering='occurence',
                                                                              reverse_chunks_order=False)
            occurences = composition_matrix.sum(0)
            occurences_ascending = is_ascending(occurences)
            self.assertTrue(occurences_ascending)

    def test_order_chunks_random(self):
        """
        Test that in 10 runs not all matrices are equal.
        :return:
        """
        done = False
        while not done:
            sets = make_random_sets(min_sets_count=9)
            chunks, composition_matrix = get_chunks_and_composition_array(sets)
            if len(chunks) == 1:
                continue
            representations_set = set()
            for i in range(10):
                _, composition_matrix = _get_ordered_chunks_and_composition_array(sets, chunks_ordering='random')
                representations_set.add(str(composition_matrix))
            self.assertGreater(len(representations_set), 1)
            done = True

    def test_order_sets_size_descending(self):
        for _ in range(1):
            sets = make_random_sets()
            chunks, composition_matrix = _get_ordered_chunks_and_composition_array(sets, sets_ordering='size')
            chunk_sizes = np.array([len(chunk) for chunk in chunks])
            set_sizes = composition_matrix.dot(chunk_sizes.reshape((-1, 1))).flatten()
            set_sizes_descending = is_ascending(set_sizes[::-1])
            self.assertTrue(set_sizes_descending)

    def test_order_sets_size_ascending(self):
        for _ in range(1):
            sets = make_random_sets()
            chunks, composition_matrix = _get_ordered_chunks_and_composition_array(sets, sets_ordering='size',
                                                                                   reverse_sets_order=False)
            chunk_sizes = np.array([len(chunk) for chunk in chunks])
            set_sizes = composition_matrix.dot(chunk_sizes.reshape((-1, 1))).flatten()
            set_sizes_ascending = is_ascending(set_sizes)
            self.assertTrue(set_sizes_ascending)

    def test_order_sets_chunk_count_descending(self):
        for _ in range(10):
            sets = make_random_sets()
            _, composition_matrix = _get_ordered_chunks_and_composition_array(sets, sets_ordering='chunk count')
            chunk_counts = composition_matrix.sum(1)
            chunk_counts_descending = is_ascending(chunk_counts[::-1])
            self.assertTrue(chunk_counts_descending)

    def test_order_sets_chunk_count_ascending(self):
        for _ in range(10):
            sets = make_random_sets()
            _, composition_matrix = _get_ordered_chunks_and_composition_array(sets, sets_ordering='chunk count',
                                                                              reverse_sets_order=False)
            chunk_counts = composition_matrix.sum(1)
            chunk_counts_ascending = is_ascending(chunk_counts)
            self.assertTrue(chunk_counts_ascending)

    def test_order_sets_random(self):
        done = False
        while not done:
            sets = make_random_sets(min_sets_count=9)
            chunks, composition_matrix = get_chunks_and_composition_array(sets)
            if len(chunks) == 1:
                continue
            representations_set = set()
            for i in range(10):
                _, composition_matrix = _get_ordered_chunks_and_composition_array(sets, sets_ordering='random')
                representations_set.add(str(composition_matrix))
            self.assertGreater(len(representations_set), 1)
            done = True


if __name__ == '__main__':
    unittest.main()
